@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 # coding: utf-8
 
-# In[2]:
+# In[1]:
 
 
 try:
@@ -12,10 +12,10 @@ except:
         pass
 
 
-# In[3]:
+# In[100]:
 
 
-version = '0.56'
+version = '0.59'
 
 import re
 import tempfile
@@ -33,7 +33,7 @@ import pickle
 
 # # Helper Functions
 
-# In[4]:
+# In[101]:
 
 
 def is_testing():
@@ -43,7 +43,7 @@ def is_testing():
         return False
 
 
-# In[5]:
+# In[102]:
 
 
 def isnotebook():
@@ -54,7 +54,7 @@ def isnotebook():
     return False
 
 
-# In[6]:
+# In[103]:
 
 
 def testPrint(*args):
@@ -64,7 +64,7 @@ def testPrint(*args):
 
 # # Requiremtns
 
-# In[7]:
+# In[104]:
 
 
 requirements = []
@@ -93,7 +93,7 @@ from IPython.core.display import display, HTML
 
 # # Strings
 
-# In[8]:
+# In[105]:
 
 
 def sFloat0(f): return "{:0.0f}".format(f)
@@ -123,49 +123,112 @@ if is_testing():
     print(sFloatSci5(1.234567))
 
 
+# # Process / Thread
+
+# In[359]:
+
+
+def mapThread(func, args, n=8):
+    from multiprocessing.pool import ThreadPool
+    tPool = ThreadPool(n)
+    res = tPool.map(func, args)
+    return res
+
+def mapProcess(func, args, n=8):
+    from multiprocessing import Pool
+    pPool = Pool(n)
+    res = pPool.map(func, args)
+    return res
+
+if is_testing():
+    print(mapProcess(str, range(10)))
+    print(mapThread(str, range(10)))
+    print(list(map(str, range(10))))
+
+
 # # Lists
 
-# In[9]:
+# In[314]:
 
 
-def listsGetPairsFromTwoLists(list1, list2):
-    return list(zip(list1, list2))
+def liChunks(l, chunk_size):
+    return [l[i:i + chunk_size] for i in range(0, len(l), chunk_size)]
 
 
-# In[10]:
+# In[315]:
 
 
 if is_testing():
-    listAbcd = ['A', 'B', 'C', 'D']
-    list1234 = ['1', '2', '3', '4']
-
-    print(listsGetPairsFromTwoLists(listAbcd, list1234))
+    print(liChunks(['a', 'b', 'c', 'd', 'e', 'f'], 1))
+    print(liChunks(['a', 'b', 'c', 'd', 'e', 'f'], 2))
+    print(liChunks(['a', 'b', 'c', 'd', 'e', 'f'], 3))
 
 
-# In[11]:
+# In[308]:
 
 
-def listSequentialOverlappingPairs(l):
-    return list(zip(l[:-1], l[1:]))
+def liSplit(a, number_of_splits):
+    idxs = [int(np.round(idx)) for idx in np.linspace(0, len(a), num=number_of_splits + 1)]
+    return [a[start:end] for start, end in zip(idxs[:-1], idxs[1:])]
 
 
-# In[12]:
+# In[309]:
 
 
 if is_testing():
-    list1234 = ['1', '2', '3', '4']
-    print(listSequentialOverlappingPairs(list1234))
+    print(liSplit(['a', 'b', 'c', 'd', 'e'], 1))
+    print(liSplit(['a', 'b', 'c', 'd', 'e'], 2))
+    print(liSplit(['a', 'b', 'c', 'd', 'e'], 3))
+    print(liSplit(['a', 'b', 'c', 'd', 'e'], 4))
+    print(liSplit(['a', 'b', 'c', 'd', 'e'], 5))
+    print(liSplit(['a', 'b', 'c', 'd', 'e'], 6))
 
 
-# In[ ]:
+# In[287]:
 
 
+def liJoin(l):
+    out = []
+    for e in l:
+        out += e
+    return out
 
 
+# In[289]:
+
+
+if is_testing():
+    print(liJoin([['a', 'b', 'c'], ['d', 'e', 'f']]))
+
+
+# In[355]:
+
+
+def liFlatten(l):
+    if isinstance(l, list) or isinstance(l, tuple):
+        out = []
+        for e in l:
+            flat = liFlatten(e)
+            if isinstance(flat, list) or isinstance(flat, tuple):
+                out += flat
+            else:
+                out.append(flat)
+        return out
+    else:
+        return l
+
+
+# In[358]:
+
+
+if is_testing():
+    print(liFlatten(['aa', ['bb', 'cc'], [['dd', ['ee', 'ff']]]]))
+    print(liFlatten([1, ['bb', 3], [['dd', ['ee', 'ff']]]]))
+
 
 # # List of Strings
 
-# In[13]:
+# In[110]:
 
 
 regex_cache = {}
@@ -175,35 +238,35 @@ def sListFilter(l, re_include=".*", re_exclude='^\b$'):
     return [e for e in l if regex_cache[re_include].search(e) and not regex_cache[re_exclude].search(e)]
 
 
-# In[14]:
+# In[111]:
 
 
 if is_testing():
     sListFilter(['good x', 'other good x', 'without the letter before y'])
 
 
-# In[15]:
+# In[112]:
 
 
 if is_testing():
     print(sListFilter(['good x', 'other good x', 'without the letter before y'], re_include='x'))
 
 
-# In[16]:
+# In[113]:
 
 
 if is_testing():
     print(sListFilter(['not enough x', 'enough xx', 'enough xx and also z'], re_include='xx+'))
 
 
-# In[17]:
+# In[114]:
 
 
 if is_testing():
     print(sListFilter(['not enough x', 'enough xx', 'enough xx and also z'], re_include='xx+', re_exclude='z'))
 
 
-# In[18]:
+# In[115]:
 
 
 def lCsv(l, force_oneline=False):
@@ -219,21 +282,21 @@ def lCsv(l, force_oneline=False):
     return ", ".join([str(e) for e in l])
 
 
-# In[19]:
+# In[116]:
 
 
 if is_testing():
     print(lCsv(['a', 'b', 1, 3.14]))
 
 
-# In[20]:
+# In[117]:
 
 
 if is_testing():
     print(lCsv([['a', 'b'], 1, 3.14]))
 
 
-# In[21]:
+# In[118]:
 
 
 if is_testing():
@@ -242,7 +305,7 @@ if is_testing():
 
 # # Text files
 
-# In[22]:
+# In[119]:
 
 
 def txtread(path):
@@ -263,7 +326,7 @@ def txtwrite(path, txt):
             return f.write(txt)
 
 
-# In[23]:
+# In[120]:
 
 
 if is_testing():
@@ -272,7 +335,7 @@ if is_testing():
 
 # # List of Strings
 
-# In[24]:
+# In[121]:
 
 
 regex_cache = {}
@@ -284,35 +347,35 @@ def sListFilter(listOfStrings, re_include=".*", re_exclude='^\b$'):
 
 # ### Examples
 
-# In[25]:
+# In[122]:
 
 
 if is_testing():
     print(sListFilter(['good x', 'other good x', 'without the letter before y']))
 
 
-# In[26]:
+# In[123]:
 
 
 if is_testing():
     print(sListFilter(['good x', 'other good x', 'without the letter before y'], re_include='x'))
 
 
-# In[27]:
+# In[124]:
 
 
 if is_testing():
     print(sListFilter(['not enough x', 'enough xx', 'enough xx and also z'], re_include='xx+'))
 
 
-# In[28]:
+# In[125]:
 
 
 if is_testing():
     print(sListFilter(['not enough x', 'enough xx', 'enough xx and also z'], re_include='xx+', re_exclude='z'))
 
 
-# In[29]:
+# In[126]:
 
 
 if is_testing():
@@ -321,7 +384,7 @@ if is_testing():
 
 # # GCP Google Cloud Platform
 
-# In[30]:
+# In[226]:
 
 
 import warnings
@@ -345,7 +408,7 @@ def _gsBlob(path):
     return blob
 
 
-# In[31]:
+# In[128]:
 
 
 def gsGsToHttp(path):
@@ -355,7 +418,7 @@ def gsHttpToGs(path):
     return path.replace("https://storage.cloud.google.com/", "gs://").split("?")[0] # "?" is for possible "?authuser=0"
 
 
-# In[32]:
+# In[375]:
 
 
 def gsLs(path):
@@ -365,7 +428,17 @@ def gsLs(path):
     return ['gs://' + bucket + '/' + blob.name for blob in client.list_blobs(prefix=prefix)]
 
 
-# In[33]:
+# In[377]:
+
+
+def gsExists(path):
+    from google.cloud import storage
+    bucket, prefix = _gsPathSplit(path)
+    client = storage.Client().bucket(bucket)
+    return client.blob(prefix).exists()
+
+
+# In[130]:
 
 
 def gsUploadFile(gsPath, path):
@@ -373,7 +446,7 @@ def gsUploadFile(gsPath, path):
     blob.upload_from_filename(path)
 
 
-# In[34]:
+# In[131]:
 
 
 def gsTxtread(path):
@@ -385,7 +458,7 @@ def gsTxtread(path):
     return blob.download_as_string().decode()
 
 
-# In[35]:
+# In[132]:
 
 
 def gsTxtwrite(path, s):
@@ -393,7 +466,7 @@ def gsTxtwrite(path, s):
     return blob.upload_from_string(s)
 
 
-# In[36]:
+# In[133]:
 
 
 def gsImwrite(path, img):
@@ -402,7 +475,7 @@ def gsImwrite(path, img):
     blob.upload_from_string(encimg.tobytes())
 
 
-# In[37]:
+# In[134]:
 
 
 def gsImread(path):
@@ -411,7 +484,7 @@ def gsImread(path):
     return cv2.imdecode(raw, 1)[:, :, [2, 1, 0]]
 
 
-# In[38]:
+# In[135]:
 
 
 def gsPklWrite(path, obj):
@@ -419,7 +492,7 @@ def gsPklWrite(path, obj):
     return blob.upload_from_string(pickle.dumps(obj, protocol=4))
 
 
-# In[39]:
+# In[136]:
 
 
 def gsPklRead(path):
@@ -429,7 +502,7 @@ def gsPklRead(path):
 
 # # Pickle
 
-# In[40]:
+# In[137]:
 
 
 def pklwrite(path, obj):
@@ -440,7 +513,7 @@ def pklwrite(path, obj):
             return pickle.dump(obj, f, protocol=4)
 
 
-# In[41]:
+# In[138]:
 
 
 def pklread(path):
@@ -451,7 +524,7 @@ def pklread(path):
             return pickle.load(f)
 
 
-# In[42]:
+# In[139]:
 
 
 if is_testing():
@@ -464,7 +537,7 @@ if is_testing():
 
 # # Images
 
-# In[43]:
+# In[140]:
 
 
 def imread(path):
@@ -498,7 +571,7 @@ def impad(img, top=0, bottom=0, left=0, right=0, color=255):
     return np.pad(img, [(top, bottom), (left, right), (0, 0)], 'constant', constant_values=color)
 
 
-# In[44]:
+# In[141]:
 
 
 def imshow(array, scale=1):
@@ -513,7 +586,7 @@ def imshow(array, scale=1):
         cv2.waitKey()
 
 
-# In[45]:
+# In[142]:
 
 
 if is_testing():
@@ -521,14 +594,14 @@ if is_testing():
     imshow(img)
 
 
-# In[46]:
+# In[143]:
 
 
 if is_testing():
     imshow('http://via.placeholder.com/70.png')
 
 
-# In[47]:
+# In[144]:
 
 
 def imscaleNN(img, s):
@@ -538,7 +611,7 @@ def imscaleBic(img, s):
     return cv2.resize(img, None, fx=s, fy=s, interpolation=cv2.INTER_CUBIC)
 
 
-# In[48]:
+# In[145]:
 
 
 if is_testing():
@@ -557,7 +630,7 @@ if is_testing():
         imshow(out_large_interbic)
 
 
-# In[49]:
+# In[146]:
 
 
 def imRepeat(img, height, width):
@@ -571,7 +644,7 @@ def imRepeat(img, height, width):
     return img
 
 
-# In[50]:
+# In[147]:
 
 
 if is_testing():
@@ -579,33 +652,33 @@ if is_testing():
     imshow(imRepeat(img_random_color_3x5, 7, 11), scale=10)
 
 
-# In[51]:
+# In[148]:
 
 
 def imNewWhite(height, width):
     return np.ones((height, width, 3)).astype(np.uint8) * 255
 
 
-# In[52]:
+# In[149]:
 
 
 if is_testing(): imshow(imNewWhite(10, 500) - 30)
 
 
-# In[53]:
+# In[150]:
 
 
 def imNewBlack(height, width):
     return np.zeros((height, width, 3)).astype(np.uint8)
 
 
-# In[54]:
+# In[151]:
 
 
 if is_testing(): imshow(imNewBlack(10, 500))
 
 
-# In[55]:
+# In[152]:
 
 
 def imAddNoiseGauss(img, std):
@@ -617,7 +690,7 @@ def imAddNoiseGauss(img, std):
 if is_testing(): imshow(imAddNoiseGauss(imread('http://via.placeholder.com/150.png'), 10))
 
 
-# In[56]:
+# In[153]:
 
 
 def imJpgDegradation(img, quality):
@@ -630,7 +703,7 @@ def imJpgDegradation(img, quality):
 if is_testing(): imshow(imJpgDegradation(imread('http://via.placeholder.com/150.png'), 15))
 
 
-# In[57]:
+# In[154]:
 
 
 def imCropCenter(img, size):
@@ -645,35 +718,35 @@ def imCropCenter(img, size):
     return img[h_start:h_end, w_start:w_end]
 
 
-# In[58]:
+# In[155]:
 
 
 if is_testing(): 
     imshow(imCropCenter(imread('http://via.placeholder.com/40.png'), 10))
 
 
-# In[59]:
+# In[156]:
 
 
 if is_testing(): 
     imshow(imCropCenter(imread('http://via.placeholder.com/40.png'), 49))
 
 
-# In[60]:
+# In[157]:
 
 
 if is_testing(): 
     imshow(imCropCenter(imread('http://via.placeholder.com/40.png'), 50))
 
 
-# In[61]:
+# In[158]:
 
 
 if is_testing(): 
     imshow(imCropCenter(imread('http://via.placeholder.com/40.png'), 100))
 
 
-# In[62]:
+# In[159]:
 
 
 def imGallery(imgs, pad=0):
@@ -704,7 +777,7 @@ def imGallery(imgs, pad=0):
     return out
 
 
-# In[63]:
+# In[160]:
 
 
 if is_testing():
@@ -712,49 +785,49 @@ if is_testing():
     listOfDummyImagesBlackToWhite = [imNewBlack(10, 10) + 25 * i for i in range(10)]
 
 
-# In[64]:
+# In[161]:
 
 
 if is_testing():
     imshow(imGallery(listOfDummyImagesBlackToWhite[:1]))
 
 
-# In[65]:
+# In[162]:
 
 
 if is_testing():
     imshow(imGallery(listOfDummyImagesBlackToWhite[:2]))
 
 
-# In[66]:
+# In[163]:
 
 
 if is_testing():
     imshow(imGallery(listOfDummyImagesBlackToWhite[:2], pad=1))
 
 
-# In[67]:
+# In[164]:
 
 
 if is_testing():
     imshow(imGallery(listOfDummyImagesBlackToWhite[:5], pad=1))
 
 
-# In[68]:
+# In[165]:
 
 
 if is_testing():
     imshow(imGallery(listOfDummyImagesBlackToWhite[:9], pad=1))
 
 
-# In[69]:
+# In[166]:
 
 
 if is_testing():
     imshow(imGallery(listOfDummyImagesBlackToWhite[:10], pad=1))
 
 
-# In[70]:
+# In[167]:
 
 
 if is_testing():
@@ -764,14 +837,14 @@ if is_testing():
 
 # # Files
 
-# In[71]:
+# In[168]:
 
 
 def fiFindByWildcard(wildcard):
     return natsort.natsorted(glob.glob(wildcard, recursive=True))
 
 
-# In[72]:
+# In[169]:
 
 
 if is_testing():
@@ -800,7 +873,7 @@ if is_testing():
         imshow(imGallery(imgs))
 
 
-# In[73]:
+# In[170]:
 
 
 def fiRemove(path):
@@ -828,7 +901,7 @@ def fiAgeHours(path):
     return (time.time() - os.stat(path)[stat.ST_MTIME]) / 3600
 
 
-# In[74]:
+# In[171]:
 
 
 if is_testing():
@@ -862,7 +935,7 @@ if is_testing():
 
 # # Jupyter
 
-# In[75]:
+# In[172]:
 
 
 def jnNbFullWidth():
@@ -876,7 +949,7 @@ def jnPandasSettings():
     pd.set_option('display.max_colwidth', None)
 
 
-# In[76]:
+# In[173]:
 
 
 def jnImageRenderingPixelated():
@@ -888,7 +961,7 @@ def jnImageRenderingAuto():
     display(HTML('<style type = text/css> img { image-rendering: auto; } </style>'))
 
 
-# In[77]:
+# In[174]:
 
 
 def jnDefault():
@@ -900,7 +973,7 @@ if is_testing():
     jnDefault()
 
 
-# In[78]:
+# In[175]:
 
 
 def jnHtmlRow(strList, width=100):
@@ -913,7 +986,7 @@ if is_testing():
     
 
 
-# In[79]:
+# In[176]:
 
 
 def jnImgToHtmlB64(img):
@@ -926,8 +999,11 @@ def jnImgToHtmlB64(img):
 #def jnHtmlImgsRow(imgs, names=None):
 
 def jnImageRow(imgs, names=None, width=100, external=False, title_img=None, title=None):
+    if isinstance(imgs, OrderedDict):
+        names, imgs = list(zip(*imgs.items()))
+        
     _row = lambda x: '<tr>' + x + '</tr>'
-    _col = lambda x: '<td style="text-align:center">' + x + '</td>'
+    _col = lambda x: '<td style="text-align:center; word-wrap:break-word;">' + x + '</td>'
     _nacols = lambda x: prefix_name + "".join(map(_col, x))
     
     prefix_name = "" if title is None else _col(title)
@@ -945,22 +1021,25 @@ def jnImageRow(imgs, names=None, width=100, external=False, title_img=None, titl
     display(HTML(html))
 
 
-# In[80]:
+# In[177]:
 
 
 if is_testing():
-    jnImageRow(imgs=[imNewBlack(20, 20), imNewBlack(40, 20), imNewBlack(20, 800)], names=["a", "b", "c"])
+    jnImageRow(imgs=[imNewBlack(20, 20), imNewBlack(40, 20), imNewBlack(20, 800)], names=["longWordNoSpace"*10, "b", "c"])
    
 
 
-# In[81]:
+# In[178]:
 
 
 if is_testing():
-    jnImageRow(imgs=["http://via.placeholder.com/40x20.png", "http://via.placeholder.com/40x20.png"], names=["a", "b"], external=True, title='test')
+    imgs = OrderedDict()
+    imgs['a'] = "http://via.placeholder.com/40x20.png"
+    imgs['b'] = "http://via.placeholder.com/40x20.png"
+    jnImageRow(imgs, external=True, title='test')
 
 
-# In[82]:
+# In[179]:
 
 
 
@@ -970,7 +1049,7 @@ if is_testing():
 
 # # Shell
 
-# In[83]:
+# In[180]:
 
 
 def shell(cmd):
@@ -985,7 +1064,7 @@ def shell(cmd):
 
 # # GIT
 
-# In[84]:
+# In[181]:
 
 
 def gitGetHash():
@@ -995,7 +1074,7 @@ if is_testing():
     print(gitGetHash())
 
 
-# In[85]:
+# In[182]:
 
 
 def gitGetBranch():
@@ -1005,7 +1084,7 @@ if is_testing():
     print(gitGetBranch())
 
 
-# In[86]:
+# In[183]:
 
 
 def gitGetMessage():
@@ -1015,7 +1094,7 @@ if is_testing():
     print(gitGetMessage())
 
 
-# In[87]:
+# In[184]:
 
 
 def gitGetAllFiles():
@@ -1025,7 +1104,7 @@ if is_testing():
     print(gitGetAllFiles())
 
 
-# In[88]:
+# In[185]:
 
 
 def gitListByTime(path=None):
@@ -1041,7 +1120,7 @@ if is_testing():
     print(gitListByTime()[-3:])
 
 
-# In[89]:
+# In[186]:
 
 
 def shellSortDirsBySize():
@@ -1053,14 +1132,14 @@ if is_testing():
 
 # # Package Files
 
-# In[90]:
+# In[187]:
 
 
 if is_testing():
     package_files = {}
 
 
-# In[91]:
+# In[188]:
 
 
 def shell(cmd):
@@ -1068,7 +1147,7 @@ def shell(cmd):
     return result.stdout.decode()
 
 
-# In[92]:
+# In[189]:
 
 
 if is_testing():
@@ -1100,7 +1179,7 @@ setup(
 )""".format(requirements=", ".join(["'" + r + "'" for r in requirements]), version=version)
 
 
-# In[93]:
+# In[190]:
 
 
 if is_testing():
@@ -1110,7 +1189,7 @@ description-file = README.md
 """
 
 
-# In[94]:
+# In[191]:
 
 
 if is_testing():
@@ -1133,14 +1212,14 @@ OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 SOFTWARE."""
 
 
-# In[95]:
+# In[192]:
 
 
 if is_testing():
     package_files['README.md'] = ""
 
 
-# In[96]:
+# In[193]:
 
 
 if is_testing():
@@ -1149,7 +1228,7 @@ if is_testing():
     
 
 
-# In[97]:
+# In[194]:
 
 
 if is_testing():
@@ -1160,7 +1239,7 @@ lx/__init__.py
 """
 
 
-# In[98]:
+# In[195]:
 
 
 if is_testing():
